;*********************************************************************************
;PUREBASIC EasyVENT.
;EASILY CREATE EVENT HANDLERS.

;Release version 3.1.2 by Stephen Rodriguez (Registered PB user: srod)
;September 2007.

;Developed in Purebasic 4.
;Windows only
;*********************************************************************************

;To use this library you simply issue the commands 
;
;     SetEventHandler() 
;and  RemoveEventHandler()
;
;to attach and revove event handlers to controls/windows etc.
;
;See the attached user guide for details.
;*********************************************************************************


EnableExplicit ;Now declare ye variables!

CompilerIf Defined(EN_LINK, #PB_Constant)
CompilerElse
  #EN_LINK = $70B
CompilerEndIf

CompilerIf Defined(ENM_LINK, #PB_Constant)
CompilerElse
  #ENM_LINK = $04000000
CompilerEndIf

CompilerIf Defined(ENLINK, #PB_Structure)
CompilerElse
Structure ENLINK
  nmhdr.NMHDR
  msg.l
  wParam.l
  lParam.l
  chrg.CHARRANGE
EndStructure
CompilerEndIf


XIncludeFile "EasyVENTResident.pbi"

;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;-Declares/prototypes
DeclareDLL.l PerformDefaultWinProcessing(*sender.PB_Sender)
DeclareDLL.l SetEventHandler(hwnd, evtype, procaddress, commanditem=-1)
Declare.l EventSetEventHandlerUtility(hwnd, evtype, procaddress, commanditem=-1)
Declare.l EventRegisterControl(hwnd)
DeclareDLL.l RemoveEventHandler(hwnd, evtype, commanditem=-1)
Declare.l EventDespatcher(hwnd, uMsg, wParam, lParam) 
Declare.l EventCallHandler(*sender.PB_Sender, result, *mess.Long, commanditem=-1)
Declare.l EventDelPropsCallback(hwnd, lpszString, hData,dwData)
Declare.l EventMouseProc(Value) 
Prototype.l CallHandler(sender.l)
; Globals
Structure EVENT_g
  gMouseThread.l
  gCurrentcontrol.l
  gDrag.l             ;0 for no drag, 1 for a possible drag instigation.
  gDragfrom.l
  gDragx.l
  gDragy.l
  gCursorCount.l
  gCursor.l
  gEnterGadget.l ;Registered message.
  gLeaveGadget.l ;Registered message.

;  gDraglist.l
;  gDragitem.l
EndStructure

Global EVENT_globals.EVENT_g
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;-Exported functions.
;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following function is called by the developer's event handlers to perform 'deafult processing'
;of the underlying Window's message (if any).
;This gives the developer a choice of performing default processing before the event handler does it's job,
;or after the event handler runs, or he/she can prevent any default processing from occurring by simply
;not calling this procedure.
ProcedureDLL.l PerformDefaultWinProcessing(*sender.PB_Sender)
  Protected result, hwnd, text$
  If *sender
    hwnd = *sender\res1
    If hwnd ; Will be zero for those events which do not carry default processing.
      text$="EVENT_oldproc"
      result = GetProp_(hwnd, @text$)
      If result 
        *sender\res2 = 1
        *sender\res3 = CallWindowProc_(result, hwnd, *sender\uMsg, *sender\wParam, *sender\lParam) 
        ProcedureReturn *sender\res3
      EndIf
    EndIf
  EndIf
  ProcedureReturn 0
EndProcedure


;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following user procedure removes an event procedure for a given window/control.
ProcedureDLL.l RemoveEventHandler(hwnd, evtype, commanditem=-1)
  Protected result, event$, hInstance,lpdwProcessId, classname$
;We determine the type of control for some individual processing.
;We use the GetClassName_() function.
  classname$=Space(100)
  GetClassName_(hwnd, @classname$, 100)
  event$="EVENT_"+Str(evtype)
  If evtype=#OnMenuItemSelect
    event$+":"+Str(commanditem) 
  ElseIf evtype=#OnKeyPress And classname$="ComboBox"
;We need to identify the handle of the edit control attached to the ComboBox.      
    result=FindWindowEx_(hwnd, #Null, "EDIT", #Null)
    If result : hwnd = result : EndIf
  EndIf
  ProcedureReturn RemoveProp_(hwnd, @event$)
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following user function registers an event procedure for a given window/control.
;Returns non-zero if no error.
ProcedureDLL.l SetEventHandler(hwnd, evtype, procaddress, commanditem=-1)
Protected result, event$, hInstance, processid, classname$
;First check hWnd is valid.
  If IsWindow_(hwnd)=#False
    ProcedureReturn 0
  EndIf
;Now we determine the type of control for some individual processing.
;We use the GetClassName_() function.
  classname$=Space(100)
  GetClassName_(hwnd, @classname$, 100)
;Now register the event as appropriate.
  Select evtype
    Case 0 To #xxx_endofstandardevents-1
      result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)

    Case #OnButtonClick
      If classname$="Button"
        result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
      EndIf    
    Case #OnChange
      If FindString(LCase(classname$),"richedit",1)
        SendMessage_(hwnd, #EM_SETEVENTMASK, 0, SendMessage_(hwnd, #EM_GETEVENTMASK, 0,0)|#ENM_CHANGE)
      EndIf
      result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
    Case #OnLinkClick
      If FindString(LCase(classname$),"richedit",1)
        SendMessage_(hwnd, #EM_SETEVENTMASK, 0, SendMessage_(hwnd, #EM_GETEVENTMASK, 0,0)|#ENM_LINK)
        result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
      EndIf    
    Case #OnDragDrop
      result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
;Now register the control for accepting dropped files.
      DragAcceptFiles_(hwnd, 1)
    Case (#xxx_endofstandardevents+1) To (#xxx_endofkeyevents-1)
      If classname$="ComboBox"
;We need to identify the handle of the edit control attached to the ComboBox.      
        result=FindWindowEx_(hwnd, #Null, "EDIT", #Null)
        If result : hwnd = result : EndIf
      EndIf
      result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
    Case #OnMenuItemSelect
;First check that a valid command item has been supplied.
      If commanditem = -1
        ProcedureReturn 0
      EndIf
      result=EventSetEventHandlerUtility(hwnd, evtype, procaddress, commanditem)
    Case #OnCollapseExpandSelection
      If FindString(LCase(classname$),"tree",1)
        result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
      EndIf
    Case #OnEditTreeLabels
      If FindString(LCase(classname$),"tree",1)
;We need to set the #TVS_EDITLABELS style.
        SetWindowLong_(hwnd, #GWL_STYLE, GetWindowLong_(hwnd, #GWL_STYLE)|#TVS_EDITLABELS)
        result=EventSetEventHandlerUtility(hwnd, evtype, procaddress)
      EndIf
    Default
      ProcedureReturn 0
  EndSelect
;Initiate the Mouse enter\exit thread to trap #OnCursorEnter and #OnCursorExit events.
  If EVENT_globals\gMouseThread = 0
    GetWindowThreadProcessId_(hwnd, @processid)
    EVENT_globals\gMouseThread= CreateThread(@EventMouseProc(), processid) 
    EVENT_globals\gEnterGadget=RegisterWindowMessage_(@"EVENT_ENTERCONTROL")
    EVENT_globals\gLeaveGadget=RegisterWindowMessage_(@"EVENT_LEAVECONTROL")
  EndIf
 ProcedureReturn result*EVENT_globals\gMouseThread
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;-Internal functions.
;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following utility procedure is called by the SetEventHandler() function to register an event handler
;and to superclass the control.
;This is used because some events are not registered with the control targeted by the user, but with
;child controls. E,g, #WM_CHAR with editable combos.
Procedure.l EventSetEventHandlerUtility(hwnd, evtype, procaddress, commanditem=-1)
  Protected result, event$, parenthWnd, anchWnd, tempAnchWnd
;First create a window property containing the address of the underlying event procedure.
  event$="EVENT_"+Str(evtype)
;Check if we are dealing with a menu/toolbar item.
  If commanditem<>-1
    event$+":"+Str(commanditem) 
  EndIf
  result=SetProp_(hwnd, @event$, procaddress)
  If result=0 : ProcedureReturn 0 : EndIf
;Now superclass the control so that we can trap windows messages.
  result*EventRegisterControl(hwnd)
  If result=0
    RemoveProp_(hwnd, @event$)
    ProcedureReturn 0
  EndIf
;Now superclass the parent control as this receives a lot of messages.
  parenthWnd=GetParent_(hwnd)
  If parenthWnd
    result*EventRegisterControl(parenthWnd)
  EndIf
;Superclass the root ancestor control if a #OnDragItemEnd is being registered.
;This allows us to trap #PB_Event_GadgetDrop and #PB_Event_WindowDrop events.
;There is a bug with GetAncestor_() under win 95 / 98 / ME.
;    anchWnd = GetAncestor_(hwnd, 2)
  If evtype = #OnDropItem
    anchWnd=hwnd 
    tempAnchWnd=GetParent_(hwnd) 
    While Not tempAnchWnd 
      anchWnd=tempAnchWnd 
      tempAnchWnd = GetParent_(tempAnchWnd) 
    Wend 
    result*EventRegisterControl(anchWnd)
  EndIf
  ProcedureReturn result
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following user function superclasses the given window/control in order that we can intercept
;messages etc. It is called by the SetEventHandler function.
;Returns non-zero if no error.
Procedure.l EventRegisterControl(hwnd)
  Protected result, event$, classname$
;Here we change the window procedure of the given window in order to examine each control
;as it is created. This allows us to set various styles; e.g. #SS_NOTIFY for static controls.
  If hwnd
    result=1 ;Signals no error at this stage of the registration.
    classname$=Space(100)
    GetClassName_(hwnd, @classname$, 100)
    If classname$="Static"
;Set the #SS_NOTIFY flag so that the static control sends mouse messages.
      SetWindowLong_(hwnd, #GWL_STYLE, GetWindowLong_(hwnd, #GWL_STYLE) | #SS_NOTIFY) 
    EndIf
;Now superclass the new control (if it hasn't already!)
    event$="EVENT_oldproc"
    If GetProp_(hwnd, @event$)=#Null
      SetProp_(hwnd, @event$, GetWindowLong_(hwnd, #GWL_WNDPROC))
      result=SetWindowLong_(hwnd, #GWL_WNDPROC, @EventDespatcher()) 
    EndIf
  EndIf
 ProcedureReturn result
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following function acts as the main 'despatcher' in that it intercepts calls to registered windows/controls
;window procedures and calls the registered event procedures for the majority of events etc.
Procedure.l EventDespatcher(hwnd, uMsg, wParam, lParam) 
  Protected result, returntrue, returnfalse, result2, text$,hInstance,lpdwProcessId, classname$,sender.PB_Sender
  Protected *nmh.NMHDR, *nmlv.NM_LISTVIEW, *pnmtv.NM_TREEVIEW, tch.TC_HITTESTINFO, *ptvdi.TV_DISPINFO
  Protected *pnmud.NM_UPDOWN, pt.POINT,hTempImgList, forehWnd, rc.RECT, *enl.ENLINK, txr.TEXTRANGE
  Protected HitTest.TV_HITTESTINFO, tvitem.TV_ITEM
  Protected font, mesent
  result = #Event_ReturnDefault ;Default.
  sender\hwnd = hwnd ;This field may change as a result of converting Win messages into EasyVENT events.
  sender\res1 = hwnd
  sender\uMsg=uMsg
  sender\wParam=wParam
  sender\lParam=lParam
  sender\button=-1
  sender\item=-1
  sender\state=-1
  sender\text$=""
  sender\res2 = 0
  sender\res3 = 0
  Select uMsg

    Case #PB_Event_GadgetDrop, #PB_Event_WindowDrop
    ;We MUST call the default proc for this to enable the Purebasic drag/drop lib to sort out it's
    ;internal structure.
      text$="EVENT_oldproc"
      result2 = GetProp_(hwnd, @text$)
      If result2 
        result= CallWindowProc_(GetProp_(hwnd, @text$), hwnd, uMsg, wParam, lParam) 
      EndIf
    ;Because the developer has no control over default processing for this message, we must prevent
    ;him/her from calling PerformDefaultWinProcessing().
      sender\res1 = 0
    If uMsg = #PB_Event_WindowDrop
      sender\hwnd = WindowID(wParam)
    Else
      sender\hwnd = GadgetID(wParam)
    EndIf
    pt\x = DesktopMouseX() : pt\y = DesktopMouseY()
    MapWindowPoints_(0, sender\hwnd, pt, 1)
    sender\mousex = pt\x
    sender\mousey = pt\y
    sender\button = #Event_LeftButton
    sender\message = #OnDropItem
    sender\item = EVENT_globals\gDragfrom
    result = EventCallHandler(@sender,result, @mesent) 
    ProcedureReturn result

    Case #WM_NCDESTROY
      ;Here we must remove all properties we have created.
      ;The following sets up a callback function which iterates through the properties.
      text$="EVENT_oldproc"
      result = GetProp_(hwnd, @text$)
      EnumPropsEx_(hwnd, @EventDelPropsCallback(),#Null)
      If result
        ProcedureReturn CallWindowProc_(result, hwnd, uMsg, wParam, lParam)
      Else
        ProcedureReturn 0
      EndIf

    Case #WM_GETMINMAXINFO 
      sender\message = #OnSizingWindow
      sender\button=#Event_LeftButton
      result = EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case EVENT_globals\gEnterGadget ;Posted by the Mouse enter\exit thread.
      ;Because there is no appropriate default processing for this message, we must prevent
      ;the developer from calling PerformDefaultWinProcessing().
        sender\res1 = 0
      sender\message = #OnCursorEnter
      sender\mousex=wParam
      sender\mousey=lParam
      EventCallHandler(@sender,result, @mesent) 
      ProcedureReturn 0 ;No default processing possible.
    
    Case EVENT_globals\gLeaveGadget ;Posted by the Mouse enter\exit thread.
      ;Because there is no appropriate default processing for this message, we must prevent
      ;the developer from calling PerformDefaultWinProcessing().
        sender\res1 = 0
      sender\message = #OnCursorExit
      sender\mousex=wParam
      sender\mousey=lParam
      EventCallHandler(@sender,result, @mesent) 
      ProcedureReturn 0 ;No default processing possible.

    Case #WM_SETCURSOR
      sender\message = #OnSetCursor
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 1 

    Case #WM_LBUTTONDOWN, #WM_NCLBUTTONDOWN
    ;In the case of a tree gadget there might well be a #OnItemCheckboxChanging handler which we
    ;have to deal with manually. Here we deal with an attempted alteration via the mouse.
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      If GetProp_(hwnd, "EVENT_"+Str(#OnItemCheckboxChanging)) ;Is there such a handler?
        classname$=Space(100)
        GetClassName_(hwnd, @classname$, 100)
        If classname$ = "SysTreeView32" ;Is it a tree gadget?
        ;Determine which item was clicked etc. 
          HitTest\pt\x = sender\mousex
          HitTest\pt\y = sender\mousey
          If SendMessage_(hwnd, #TVM_HITTEST, 0, HitTest) And HitTest\flags = #TVHT_ONITEMSTATEICON
            ;Identify the zero-based item index of the item clicked.
              tvitem\mask = #TVIF_PARAM	| #TVIF_STATE
              tvitem\stateMask = #TVIS_STATEIMAGEMASK
              tvitem\hItem = HitTest\hItem
              SendMessage_(hwnd, #TVM_GETITEM, 0, tvitem)
            sender\item = tvitem\lParam
            sender\message = #OnItemCheckboxChanging
            ;Because there is no appropriate default processing for this message, we must prevent
            ;the developer from calling PerformDefaultWinProcessing().
              sender\res1 = 0
            ;Set sender\state to #EVENT_UNCHECKING or #EVENT_CHECKING as appropriate.
              sender\state = tvitem\state>>12
            result=EventCallHandler(@sender,result, @mesent) 
          EndIf
        EndIf
      EndIf
      If result = #Event_ReturnDefault Or result = #Event_ReturnTrue
        mesent = 0
        sender\message=#OnMouseDown
        ;Now call any #OnMouseDown handler.
          sender\res1 = hwnd ;Allow the user to call the default window.
          sender\button=#Event_LeftButton
          sender\state = #EVENT_CLIENT
          If uMsg = #WM_NCLBUTTONDOWN
            sender\state = #EVENT_NONCLIENT
          EndIf
          result = EventCallHandler(@sender,result, @mesent) 
      EndIf
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      ;The user could be attempting a drag operation.
      ;We proceed only if the gadget is registered for drags.
      ;This avoids unnecessary use of SetCapture_() which can cause problems.
        If GetProp_(hwnd,"EVENT_"+Str(#OnDragItemStart)) And uMsg = #WM_LBUTTONDOWN
          EVENT_globals\gDrag = 1
          EVENT_globals\gDragfrom = hwnd
          EVENT_globals\gDragx = lParam&$FFFF
          EVENT_globals\gDragy = lParam>>16&$FFFF
          SetCapture_(hwnd)
        EndIf

    Case #WM_MBUTTONDOWN, #WM_NCMBUTTONDOWN
      sender\message=#OnMouseDown
      sender\button=#EVENT_MIDDLEBUTTON
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCMBUTTONDOWN
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_RBUTTONDOWN, #WM_NCRBUTTONDOWN
      sender\message=#OnMouseDown
      sender\button=#EVENT_RIGHTBUTTON
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCRBUTTONDOWN
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_LBUTTONUP, #WM_NCLBUTTONUP
butup:
      If EVENT_globals\gDrag
        EVENT_globals\gDrag = 0
        ReleaseCapture_()
      EndIf
      sender\button=#Event_LeftButton
      sender\message=#OnMouseUp
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCLBUTTONUP
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_MBUTTONUP, #WM_NCMBUTTONUP
      sender\message=#OnMouseUp
      sender\button=#EVENT_MIDDLEBUTTON
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCMBUTTONUP
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_RBUTTONUP, #WM_NCRBUTTONUP
      sender\message=#OnMouseUp
      sender\button=#EVENT_RIGHTBUTTON
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCRBUTTONUP
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_MOUSEMOVE, #WM_NCMOUSEMOVE
      If EVENT_globals\gDrag
        ;Check mousebutton. It is possible that the button is up in a ListIcon.
        If GetAsyncKeyState_(#VK_LBUTTON)&32768<>32768
          Goto butup
        EndIf
        ;Check whether the mouse has moved enough to instigate a drag.
          SetRect_(rc, EVENT_globals\gDragx-15, EVENT_globals\gDragy-15,EVENT_globals\gDragx+15,EVENT_globals\gDragy+15)
          pt\x = lParam&$FFFF
          pt\y = lParam>>16&$FFFF
          If PtInRect_(rc,pt\x,pt\y)=0
            EVENT_globals\gDrag = 0
            ReleaseCapture_()
            ;Because there is no appropriate default processing for this message, we must prevent
            ;the developer from calling PerformDefaultWinProcessing().
              sender\res1 = 0
            ;Send a #OnDragItemStart message.
              sender\message=#OnDragItemStart
              sender\button=#Event_LeftButton
              sender\mousex = pt\x : sender\mousey = pt\y
              EventCallHandler(@sender,result, @mesent) 
              ProcedureReturn 0
          EndIf
      Else
        sender\message=#OnMouseOver
        sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
        sender\state = #EVENT_CLIENT
        If uMsg = #WM_NCMOUSEMOVE
          sender\state = #EVENT_NONCLIENT
        EndIf
        result=EventCallHandler(@sender,result, @mesent) 
        ;Default return values.
          returntrue = 0 : returnfalse = 0 
      EndIf

    Case #WM_LBUTTONDBLCLK, #WM_NCLBUTTONDBLCLK
      sender\message=#OnDblClick
      sender\button=#Event_LeftButton
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCLBUTTONDBLCLK
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_RBUTTONDBLCLK, #WM_NCRBUTTONDBLCLK
      sender\message=#OnDblClick
      sender\button=#EVENT_RIGHTBUTTON
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCRBUTTONDBLCLK
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_MBUTTONDBLCLK, #WM_NCMBUTTONDBLCLK
      sender\message=#OnDblClick
      sender\button=#EVENT_MIDDLEBUTTON
      sender\mousex = lParam&$FFFF : sender\mousey = lParam>>16&$FFFF
      sender\state = #EVENT_CLIENT
      If uMsg = #WM_NCMBUTTONDBLCLK
        sender\state = #EVENT_NONCLIENT
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_COMMAND
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      result2 = wParam>>16 & $FFFF
      Select result2
        Case #BN_CLICKED
          If lParam ;Button
;Now we determine the type of control to determine if it is a toolbar.
;We use the GetClassName_() function.
            classname$=Space(100)
            GetClassName_(lParam, @classname$, 100)
            If classname$ = "ToolbarWindow32" : Goto toolbar : EndIf
            sender\hwnd = lParam ;Handle of button.
            sender\message = #OnButtonClick
            sender\button=#Event_LeftButton
            sender\item=wParam&$FFFF
            result=EventCallHandler(@sender,result, @mesent) 
          Else ;Menu item.
toolbar:    sender\message = #OnMenuItemSelect
            sender\button=#Event_LeftButton
            sender\item=wParam&$FFFF
            result=EventCallHandler(@sender,result, @mesent,wParam&$FFFF) 
          EndIf
        Case #EN_CHANGE
            sender\hwnd = lParam ;Handle of Edit control.
            sender\message = #OnChange
            result=EventCallHandler(@sender,result, @mesent) 
        Case #LBN_SELCHANGE ;Same as #CBN_SELCHANGE
          If lParam
            sender\hwnd = lParam ;Handle of ListBox.
            sender\message = #OnItemSelected
            result=EventCallHandler(@sender,result, @mesent) 
          Else ;Keyboard accelerator. Same value as #LBN_SELCHANGE = 1 !
            Goto toolbar
          EndIf
        Case #LBN_SELCANCEL ;Same as #CBN_SETFOCUS
          If lParam
            ;Check listbox.
              classname$=Space(100)
              GetClassName_(lParam, @classname$, 100)
            If classname$ = "ListBox"
              sender\hwnd = lParam ;Handle of ListBox.
              sender\message = #OnItemSelected
              result=EventCallHandler(@sender,result, @mesent) 
            EndIf
          EndIf
      EndSelect

    Case #WM_SYSCOMMAND 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      Select wParam &$FFF0
        Case #SC_MAXIMIZE 
          sender\message = #OnMaximize
          result=EventCallHandler(@sender,result, @mesent) 
        Case #SC_MINIMIZE 
          sender\message = #OnMinimize
          result=EventCallHandler(@sender,result, @mesent) 
      EndSelect 

    Case #WM_NOTIFY
      *nmh=lParam
      Select *nmh\code
        Case #EN_LINK
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          *enl=lParam
          If *enl\msg=#WM_LBUTTONDOWN Or *enl\msg=#WM_RBUTTONDOWN
            sender\hwnd = *enl\nmhdr\hwndFrom ;Handle of Rich edit control.
            sender\message = #OnLinkClick
            If *enl\msg=#WM_LBUTTONDOWN
              sender\button=#Event_LeftButton
            Else
              sender\button=#EVENT_RIGHTBUTTON
            EndIf
            sender\item=*enl\chrg\cpMin
            sender\state=*enl\chrg\cpMax
            sender\text$=Space(sender\state-sender\item+2)
            txr\chrg\cpMin=*enl\chrg\cpMin
            txr\chrg\cpMax=*enl\chrg\cpMax
            txr\lpstrText=@sender\text$
            SendMessage_(sender\hwnd,#EM_GETTEXTRANGE,0,txr)
            result=EventCallHandler(@sender,result, @mesent) 
          EndIf

        Case #LVN_ITEMCHANGING, #LVN_ITEMCHANGED ;ListIcons and ExplorerList gadgets.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          *nmlv=lParam
          sender\hwnd = *nmlv\hdr\hwndFrom ;Handle of ListIcon.
          sender\message = #OnItemSelected
          If *nmh\code = #LVN_ITEMCHANGING
            sender\message = #OnItemSelecting
          EndIf
          sender\item = *nmlv\iItem
          If *nmlv\uOldState&#LVIS_SELECTED=#LVIS_SELECTED
            sender\state=#EVENT_DESELECT
            result=EventCallHandler(@sender,result, @mesent) 
          ElseIf *nmlv\uNewState&#LVIS_SELECTED=#LVIS_SELECTED
            sender\state=#EVENT_SELECT
            result=EventCallHandler(@sender,result, @mesent) 
          ;NOTE to apply the following to tree gadgets requires #TVN_ITEMCHANGED which is only valid
          ;on Win Vista.
          ElseIf *nmlv\uOldState& #LVIS_STATEIMAGEMASK <> 0 And *nmlv\uNewState& #LVIS_STATEIMAGEMASK And *nmh\code = #LVN_ITEMCHANGING
            sender\message = #OnItemCheckboxChanging
            ;Set sender\state to #EVENT_UNCHECKING or #EVENT_CHECKING as appropriate.
            sender\state = 3-(*nmlv\uNewState& #LVIS_STATEIMAGEMASK)>>12
            result=EventCallHandler(@sender,result, @mesent) 
          EndIf

        Case #TVN_ITEMEXPANDING ;TreeGadgets and ExplorerTrees.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          *pnmtv=lParam
          sender\hwnd = *pnmtv\hdr\hwndFrom ;Handle of TreeGadget.
          sender\message = #OnCollapseExpandSelection
          sender\item = *pnmtv\itemNew\lParam
          sender\state = *pnmtv\action
          result=EventCallHandler(@sender,result, @mesent) 

        Case #TVN_SELCHANGING ;TreeGadgets and ExplorerTrees.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          *pnmtv=lParam
          sender\hwnd = *pnmtv\hdr\hwndFrom ;Handle of TreeGadget.
          sender\message = #OnItemSelecting
          sender\item = *pnmtv\itemNew\lParam
          sender\state = *pnmtv\action ; #TVC_BYKEYBOARD or #TVC_BYMOUSE or #TVC_UNKNOWN.
          result=EventCallHandler(@sender,result, @mesent) 

        Case #TVN_SELCHANGED ;TreeGadgets and ExplorerTrees.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          *pnmtv=lParam
          sender\hwnd = *pnmtv\hdr\hwndFrom ;Handle of TreeGadget.
          sender\message = #OnItemSelected
          sender\item = *pnmtv\itemNew\lParam
          sender\state = *pnmtv\action ; #TVC_BYKEYBOARD or #TVC_BYMOUSE or #TVC_UNKNOWN.
          result=EventCallHandler(@sender,result, @mesent) 

        Case #TVN_BEGINLABELEDIT ;Editing labels in a tree gadget.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          *ptvdi=lParam
          sender\hwnd = *ptvdi\hdr\hwndFrom ;Handle of TreeGadget.
          sender\message = #OnEditTreeLabels
          sender\item = *ptvdi\item\lParam
          sender\state = #EVENT_BEGINLABELEDIT
          result=EventCallHandler(@sender,result, @mesent) 

        Case #TVN_ENDLABELEDIT ;Editing labels in a tree gadget. 
          ;Default return values.
            returntrue = 1 : returnfalse = 0 
          *ptvdi=lParam 
          sender\hwnd = *ptvdi\hdr\hwndFrom ;Handle of TreeGadget. 
          sender\message = #OnEditTreeLabels 
          sender\item = *ptvdi\item\lParam 
          sender\state = #EVENT_ENDLABELEDIT 
          If *ptvdi\item\pszText <> #Null 
             sender\text$=PeekS(*ptvdi\item\pszText) 
          EndIf 
          result=EventCallHandler(@sender,result, @mesent) 

        Case #TCN_SELCHANGE ;Panel gadgets.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          ;Call default proc. as the panel gadget misbehaves otherwise!
            text$="EVENT_oldproc"
            result = GetProp_(hwnd, @text$)
            If result 
              CallWindowProc_(result, hwnd, uMsg, wParam, lParam) 
            EndIf
          sender\hwnd = *nmh\hwndFrom ;Handle of Tab control.
          sender\message = #OnItemSelected
;Identify which tab was selected.
          If GetAsyncKeyState_(#VK_LBUTTON)&32768 ;Selection by mouse button.
            GetCursorPos_(tch\pt)
            MapWindowPoints_(#Null,sender\hwnd,tch\pt,1)
            sender\item = SendMessage_(sender\hwnd, #TCM_HITTEST,0, tch)
            sender\state = #EVENT_MOUSE
            result=EventCallHandler(@sender,result, @mesent) 
          Else
            sender\item = SendMessage_(sender\hwnd, # TCM_GETCURSEL,0,0)
            sender\state = #EVENT_KEYBOARD
            result=EventCallHandler(@sender,result, @mesent) 
          EndIf
          ProcedureReturn 0

        Case #TCN_SELCHANGING ;Panel gadgets.
          ;Default return values.
            returntrue = 0 : returnfalse = 1 
          sender\hwnd = *nmh\hwndFrom ;Handle of Tab control.
          sender\message = #OnItemSelecting
          ;Identify which tab was selected.
          sender\state = #EVENT_KEYBOARD
          If GetAsyncKeyState_(#VK_LBUTTON)&32768 ;Selection by mouse button.
            GetCursorPos_(tch\pt)
            MapWindowPoints_(#Null,sender\hwnd,tch\pt,1)
            sender\item = SendMessage_(sender\hwnd, #TCM_HITTEST,0, tch)
            sender\state = #EVENT_MOUSE
            result=EventCallHandler(@sender,result, @mesent) 
          ElseIf GetAsyncKeyState_(#VK_LEFT)&32768 And SendMessage_(sender\hwnd, #TCM_GETCURSEL,0,0)>0;Selection by left cursor.
            sender\item = SendMessage_(sender\hwnd, #TCM_GETCURSEL,0,0)-1
            result=EventCallHandler(@sender,result, @mesent) 
          ElseIf GetAsyncKeyState_(#VK_RIGHT)&32768 And SendMessage_(sender\hwnd, #TCM_GETCURSEL,0,0)<SendMessage_(sender\hwnd, #TCM_GETITEMCOUNT,0,0)-1;Selection by right cursor.
            sender\item = SendMessage_(sender\hwnd, #TCM_GETCURSEL,0,0)+1
            result=EventCallHandler(@sender,result, @mesent) 
          ElseIf GetAsyncKeyState_(#VK_END)&32768 And SendMessage_(sender\hwnd, #TCM_GETCURSEL,0,0)<SendMessage_(sender\hwnd, #TCM_GETITEMCOUNT,0,0)-1;Selection by End button.
            sender\item = SendMessage_(sender\hwnd, #TCM_GETITEMCOUNT,0,0)-1
            result=EventCallHandler(@sender,result, @mesent) 
          ElseIf GetAsyncKeyState_(#VK_HOME)&32768 And SendMessage_(sender\hwnd, #TCM_GETCURSEL,0,0)>0;Selection by Home button.
            sender\item = 0
            result=EventCallHandler(@sender,result, @mesent) 
          EndIf
          If result = #Event_ReturnTrue Or result = #Event_ReturnDefault
            ;Call default proc. as the panel gadget misbehaves otherwise!
              text$="EVENT_oldproc"
              result2 = GetProp_(hwnd, @text$)
              If result2 
                ProcedureReturn CallWindowProc_(result2, hwnd, uMsg, wParam, lParam) 
              EndIf
          EndIf
      EndSelect

    Case #WM_CHAR
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnKeyPress
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_CLOSE
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnClose
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_CONTEXTMENU
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnContextmenuPopup
      sender\button=#EVENT_RIGHTBUTTON
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_DROPFILES 
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnDragDrop
      sender\button=#Event_LeftButton
      sender\item=DragQueryFile_(wParam , -1, 0, 0) ;Number if files dropped.
      result=EventCallHandler(@sender,result, @mesent) 
      ;Release the memory from the underlying windows structure.
      DragFinish_(wParam) 

    Case #WM_KEYDOWN
    ;In the case of a tree gadget there might well be a #OnItemCheckboxChanging handler which we
    ;have to deal with manually. Here we deal with a push of the spacebar.
      If GetProp_(hwnd, "EVENT_"+Str(#OnItemCheckboxChanging)) And wParam = #VK_SPACE ;Is there such a handler?
        classname$=Space(100)
        GetClassName_(hwnd, @classname$, 100)
        If classname$ = "SysTreeView32" ;Is it a tree gadget?
          ;Identify the state of the checkbox.
            tvitem\mask = #TVIF_PARAM	| #TVIF_STATE
            tvitem\stateMask = #TVIS_STATEIMAGEMASK
            tvitem\hItem = SendMessage_(hwnd, #TVM_GETNEXTITEM, #TVGN_CARET, 0)
            SendMessage_(hwnd, #TVM_GETITEM, 0, tvitem)
            sender\item = tvitem\lParam
            sender\message = #OnItemCheckboxChanging
            ;Because there is no appropriate default processing for this message, we must prevent
            ;the developer from calling PerformDefaultWinProcessing().
              sender\res1 = 0
            ;Set sender\state to #EVENT_UNCHECKING or #EVENT_CHECKING as appropriate.
              sender\state = tvitem\state>>12
            result=EventCallHandler(@sender,result, @mesent) 
        EndIf
      EndIf
      If result = #Event_ReturnDefault Or result = #Event_ReturnTrue
        mesent = 0
        sender\message = #OnKeyDown
        sender\res1 = hwnd ;Allow the user to call the default window.
        result=EventCallHandler(@sender,result, @mesent) 
      EndIf
      ;Default return values.
        returntrue = 0 : returnfalse = 0 

    Case #WM_KEYUP
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnKeyUp
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_KILLFOCUS
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnLostFocus
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_MOVE
      ;Default return values.
        returntrue = 0 : returnfalse = 0 
      sender\message = #OnMove
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_PAINT, #WM_NCPAINT
      ;Default return values.
        returntrue = 1 : returnfalse = 0 
      sender\message = #OnPaint
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_ERASEBKGND 
      ;Default return values.
        returntrue = 0 : returnfalse = 1 
      sender\message = #OnErase
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_SETFOCUS
      ;Default return values.
        returntrue = 0 : returnfalse = 1 
      sender\message = #OnGotFocus
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_SIZE
      ;Default return values.
        returntrue = 0 : returnfalse = 1 
      sender\message = #OnResize
      result=EventCallHandler(@sender,result, @mesent) 

    Case #WM_VSCROLL, #WM_HSCROLL
      ;Default return values.
        returntrue = 0 : returnfalse = 1 
      sender\message = #OnScroll
      sender\item = wParam>>16&$FFFF ;Scroll box position.
      sender\state = wParam&$FFFF ;Scroll bar code.
      ;Now we determine the type of control to determine if it is a SpinGadget.
      ;We use the GetClassName_() function.
      classname$=Space(100)
      GetClassName_(lParam, @classname$, 100)
      If classname$ = "msctls_updown32"
      ;We need to identify the 'buddy' window (this will be an edit control for Purebasic).
        sender\hwnd=SendMessage_(lParam, #UDM_GETBUDDY, 0,0)
      ElseIf lParam ;Some other scroll bar control
        sender\hwnd=lParam
      EndIf
      result=EventCallHandler(@sender,result, @mesent) 
  EndSelect

  If mesent = 0 ;Message hasn't been handled.
    sender\hwnd=hwnd
    sender\originalmessage=sender\message
    sender\message = #OnUnhandledWinMessage
    result=EventCallHandler(@sender,result, @mesent) 
    If mesent ;Message was handled.
      ProcedureReturn result ;In this case the developer can set any return value.
    EndIf
  EndIf

  If mesent = 0 ;Message still hasn't been handled.
    PerformDefaultWinProcessing(sender)
    result = #Event_ReturnDefault
  EndIf

;Sort out a return value.
  Select result
    Case #Event_ReturnDefault
      If sender\res2 = #True
        ProcedureReturn sender\res3
      Else
        ProcedureReturn returntrue  ;This would have been set up by the appropriate 'Case' statement above.
      EndIf
    Case #Event_ReturnFalse
        ProcedureReturn returnfalse  ;This would have been set up by the appropriate 'Case' statement above.
    Default ;#Event_ReturnTrue
        ProcedureReturn returntrue  ;This would have been set up by the appropriate 'Case' statement above.
  EndSelect
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following utility function is used by the EventDespatcher() function to call an
;event handler for the given message (if one exists).
Procedure.l EventCallHandler(*sender.PB_Sender, result, *mess.Long, commanditem=-1)
  Protected Handler.CallHandler, pt.POINT, text$
;If we are dealing with a mouse message, then we do not need to calculate the cursor position.
  If *sender\message > #OnMouseUp
    GetCursorPos_(pt)
    MapWindowPoints_(#Null,*sender\hwnd,pt,1)
    *sender\mousex = pt\x
    *sender\mousey = pt\y
  EndIf
;Determine if the control has registered a #OnANYevent.
;We discount such a handler if the underlying message is #OnUnhandledWindMessage.
  If *sender\message <> #OnUnhandledWinMessage
    text$="EVENT_"+Str(#OnANYevent)
    Handler = GetProp_(*sender\hwnd, @text$)
  EndIf
  If Handler = 0 ;No #OnANYevent handler.
  ;Instead, determine if the control has registered the particular event.
    text$="EVENT_"+Str(*sender\message)
    If *sender\message=#OnMenuItemSelect
      text$+":"+Str(commanditem) 
    EndIf
    Handler = GetProp_(*sender\hwnd, @text$)
  EndIf
  If Handler
    result=Handler(*sender) 
    If *mess
      *mess\l=1
    EndIf
    If *sender\message=#OnDragDrop
;Release the memory from the underlying windows structure.
      DragFinish_(*sender\wParam) 
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following callback function is called by windows as a result of the EnumPropsEx_() function
;issued when a registered control is destroyed.
;We use this to delete the properties we have created.
Procedure.l EventDelPropsCallback(hwnd, lpszString, hData,dwData)
  Protected text$
  If lpszString>>16<>0 ;Confirms that this parameter points to a string and is not merely an atom.
    text$ = PeekS(lpszString)
    If Left(PeekS(lpszString),6)="EVENT_"
      RemoveProp_(hwnd, lpszString)
    EndIf
    If text$ = "EVENT_oldproc"
      SetWindowLong_(hwnd, #GWL_WNDPROC, hData)
    EndIf
  EndIf
ProcedureReturn 1  
EndProcedure
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


;-#OnCursorEnter / #OnCursorExit
;//////////////////////////////////////////////////////////////////////////////////////////////////////////
;The following function acts as the thread procedure dealing with (and despatches) the #OnCursorEnter
;and #OnCursorExit events.
Procedure.l EventMouseProc(processid) 
  Protected window, child, state, pt.POINT, pid
  Repeat
    GetCursorPos_(pt)
    window=WindowFromPoint_(pt\x,pt\y)
;Check for the occurrence of OnCursorEnter or OnCursorExit events.
    If window <> EVENT_globals\gCurrentcontrol
      If EVENT_globals\gCurrentcontrol ;Exiting a control.
        PostMessage_(EVENT_globals\gCurrentcontrol,EVENT_globals\gLeaveGadget,-1,-1) ;Mouse coords have no meaning. 
      EndIf
;Entering a control.
      GetWindowThreadProcessId_(window, @pid)
      If pid=processid
        MapWindowPoints_(#Null, window, pt,1) ;Convert screen coordinates to client coordinates.
        EVENT_globals\gCurrentcontrol = window
        PostMessage_(window,EVENT_globals\gEnterGadget,pt\x,pt\y) 
      Else
        EVENT_globals\gCurrentcontrol=0
      EndIf
    EndIf
  Delay(1)
  ForEver
EndProcedure 
;//////////////////////////////////////////////////////////////////////////////////////////////////////////


; IDE Options = PureBasic 4.20 Beta 1 (Windows - x86)
; CursorPosition = 609
; FirstLine = 571
; Folding = --