;EasyVENT demonstration program - ListIcon gadgets.
;*********************************************************************************

;(Checked working with EasyVENT 3.0.0)

;DEMONSTRATES THE FOLLOWING EVENTS:
;   #OnDragDrop, #OnItemSelected and #OnItemCheckboxChanging.
;This short demo shows how to use EasyVENT to quickly receive dropped files and place
;their names into a ListIcon gadget. It also shows how to trap when a user selects 
;or (separately) deselects items in a ListIcon or checks/clears a ListIcon checkmark etc.
;*********************************************************************************


;ESSENTIAL INCLUDE.
;*********************************************************************************
XIncludeFile "EasyVENT.pbi"
;*********************************************************************************

Declare.l DragDrop(*sender.PB_Sender)
Declare.l SelectItem(*sender.PB_Sender)
Declare.l CheckBoxChange(*sender.PB_Sender)

Define Event.l, text$

OpenWindow(0, 100, 100, 350, 250, "EasyVENT ListIconGadget demo.", #PB_Window_SystemMenu |#PB_Window_MinimizeGadget|#PB_Window_MaximizeGadget| #PB_Window_ScreenCentered)
  CreateGadgetList(WindowID(0))
  TextGadget(0, 10, 10, 200, 20, "Drop some files onto the ListIcon gadget...")
  ListIconGadget(1,10,40,330,150,"Dropped files.",300,#PB_ListIcon_MultiSelect|#PB_ListIcon_CheckBoxes) 
  text$="Select and deselect items!"+Chr(10)+Chr(13)+Chr(10)+Chr(13)
  TextGadget(2, 10, 200, 200, 50, text$)
 
;Set event handlers.
  SetEventHandler(GadgetID(1), #OnDragDrop, @DragDrop())
  SetEventHandler(GadgetID(1), #OnItemSelected, @SelectItem())
  SetEventHandler(GadgetID(1), #OnItemCheckboxChanging, @CheckBoxChange())
  
Repeat
  Event = WaitWindowEvent()
Until Event = #PB_Event_CloseWindow
End


;*********************************************************************************
;                                 EVENT HANDLERS
;*********************************************************************************

;*sender\item holds the number of filenames dropped.
;*sender\wParam holds the windows handle of the underlying drop structure.
Procedure.l DragDrop(*sender.PB_Sender)
  Protected numfiles, i, buffer$
  numfiles=*sender\item
  For i = 0 To numfiles - 1 
;Get next filename.
    buffer$=Space(DragQueryFile_(*sender\wParam, i, 0, 0)+1) ;This will hold individual filenames.
    DragQueryFile_ (*sender\wParam, i, buffer$, Len(buffer$)) 
;Now add filename to the ListIcon gadget.
    AddGadgetItem(1, -1, buffer$) 
  Next 
  ProcedureReturn #Event_ReturnTrue ;Return not important.
EndProcedure


;The following '#OnSelectItem' handler is called whenever the underlying ListIcon selection changes.
;This is thus called whenever items are deselected as well as selected.
;The 'item' field tells us which item is being affected and the 'state' field tells us whether
;the item is being deselected or selected.
Procedure.l SelectItem(*sender.PB_Sender)
  Select *sender\state
    Case #EVENT_DESELECT
      Debug "Deselecting: " + Str(*sender\item)
    Case #EVENT_SELECT
      Debug "Selecting: " + Str(*sender\item)
  EndSelect
  ProcedureReturn #Event_ReturnTrue ;Return not important.
EndProcedure


;The following '#OnItemCheckboxChanging' handler is called whenever a checkbox in the underlying 
;ListIcon changes through user interaction.
;Return #Event_ReturnFalse to prevent the alteration.
Procedure.l CheckBoxChange(*sender.PB_Sender)
  Select *sender\state
    Case #EVENT_UNCHECKING
      Debug "Un-checking: " + Str(*sender\item)
    Case #EVENT_CHECKING
      Debug "Checking: " + Str(*sender\item)
  EndSelect
ProcedureReturn #Event_ReturnTrue
EndProcedure

; IDE Options = PureBasic 4.10 (Windows - x86)
; CursorPosition = 35
; FirstLine = 13
; Folding = -